/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthm@freeswitch.org>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Anthony Minessale II <anthm@freeswitch.org>
 * Dragos Oancea <dragos@freeswitch.org>
 *
 * switch_jitterbuffer.c -- Audio/Video Jitter Buffer
 *
 */
#include <switch.h>
#include <switch_jitterbuffer.h>
#include "private/switch_hashtable_private.h"

#define NACK_TIME 80000
#define RENACK_TIME 100000
#define MAX_FRAME_PADDING 2
#define MAX_MISSING_SEQ 20

/* Max size of the Redundant Encoding hash. */
#define SWITCH_RED_MAX_BUF_SZ 50 /* Keep as many redundancy packets (raw FEC) */
#define RTP_MATERIAL_LIM ULP_LVL_GROUP_SIZE_MAX * 2 /*FEC RX: ulpfec rfc5109, 48 * 2 */

/* Max diff of seq in hash, sequences older than this (counting from the most recently inserted one) get removed from hash to prevent it from growing indefinitely. */
#define SWITCH_RED_MAX_SEQ_DIFF 1400
//#define DEBUG_RED 1
#define DEBUG_ULPFEC_JB 0
#define jb_has_fec(_jb) (_jb->ulp_policy && _jb->ulp_policy->user_data)

#if DEBUG_ULPFEC_JB
#define jb_debug(_jb, _level, _format, ...) switch_log_printf(SWITCH_CHANNEL_SESSION_LOG_CLEAN(_jb->session), SWITCH_LOG_ALERT, "JB:%p:%s:%d/%d lv:%d ln:%.4d sz:%.3u/%.3u/%.3u/%.3u c:%.3u %.3u/%.3u/%.3u/%.3u %.2f%% ->" _format, (void *) _jb, (jb->type == SJB_TEXT ? "txt" : (jb->type == SJB_AUDIO ? "aud" : "vid")), _jb->allocated_nodes, _jb->visible_nodes, _level, __LINE__,  _jb->min_frame_len, _jb->max_frame_len, _jb->frame_len, _jb->complete_frames, _jb->period_count, _jb->consec_good_count, _jb->period_good_count, _jb->consec_miss_count, _jb->period_miss_count, _jb->period_miss_pct, __VA_ARGS__)
#else
#define jb_debug(_jb, _level, _format, ...) if (_jb->debug_level >= _level) switch_log_printf(SWITCH_CHANNEL_SESSION_LOG_CLEAN(_jb->session), SWITCH_LOG_ALERT, "JB:%p:%s:%d/%d lv:%d ln:%.4d sz:%.3u/%.3u/%.3u/%.3u c:%.3u %.3u/%.3u/%.3u/%.3u %.2f%% ->" _format, (void *) _jb, (jb->type == SJB_TEXT ? "txt" : (jb->type == SJB_AUDIO ? "aud" : "vid")), _jb->allocated_nodes, _jb->visible_nodes, _level, __LINE__,  _jb->min_frame_len, _jb->max_frame_len, _jb->frame_len, _jb->complete_frames, _jb->period_count, _jb->consec_good_count, _jb->period_good_count, _jb->consec_miss_count, _jb->period_miss_count, _jb->period_miss_pct, __VA_ARGS__)
#endif
//const char *TOKEN_1 = "ONE";
//const char *TOKEN_2 = "TWO";

struct switch_jb_s;

typedef struct switch_jb_node_s {
	struct switch_jb_s *parent;
	switch_rtp_packet_t packet;
	uint32_t len;
	uint8_t visible;
	uint8_t bad_hits;
	struct switch_jb_node_s *prev;
	struct switch_jb_node_s *next;
	/* used for counting the number of partial or complete frames currently in the JB */
	switch_bool_t complete_frame_mark;
} switch_jb_node_t;

struct switch_jb_s {
	struct switch_jb_node_s *node_list;
	uint32_t last_target_seq;
	uint32_t highest_read_ts;
	uint32_t highest_dropped_ts;
	uint32_t highest_read_seq;
	uint32_t highest_wrote_ts;
	uint16_t highest_wrote_seq;
	uint16_t target_seq;
	uint32_t target_ts;
	uint32_t last_target_ts;
	uint16_t psuedo_seq;
	uint16_t last_psuedo_seq;
	uint32_t visible_nodes;
	uint32_t allocated_nodes;
	uint32_t complete_frames;
	uint32_t frame_len;
	uint32_t min_frame_len;
	uint32_t max_frame_len;
	uint32_t highest_frame_len;
	uint32_t period_miss_count;
	uint32_t consec_miss_count;
	uint32_t period_miss_inc;
	double period_miss_pct;
	uint32_t period_good_count;
	uint32_t consec_good_count;
	uint32_t period_count;
	uint32_t dropped;
	uint32_t samples_per_frame;
	uint32_t samples_per_second;
	uint32_t bitrate_control;
	uint32_t video_low_bitrate;
	uint8_t write_init;
	uint8_t read_init;
	uint8_t debug_level;
	uint16_t next_seq;
	switch_size_t last_len;
	switch_inthash_t *missing_seq_hash;
	switch_inthash_t *node_hash;
	switch_inthash_t *node_hash_ts;
	switch_mutex_t *mutex;
	switch_mutex_t *list_mutex;
	switch_memory_pool_t *pool;
	int free_pool;
	int drop_flag;
	switch_jb_flag_t flags;
	switch_jb_type_t type;
	switch_core_session_t *session;
	switch_channel_t *channel;
	uint32_t buffer_lag;
	uint32_t flush;
	uint32_t packet_count;
	uint32_t max_packet_len;
	uint32_t period_len;
	uint32_t nack_saved_the_day;
	uint32_t nack_didnt_save_the_day;

	/* RED ulpfec */
	ulp_policy_t *ulp_policy;					/* ULP context. */

	switch_inthash_t	*redundancy_hash;		/* Hash of the Redundant Encodings. Hashed by Redundant Encoding's key. For ULPFEC scheme it is FEC's snb (network ordered). */
	switch_red_t		*redundancy_first;		/* pointer to the oldest Redundant Encoding packet in the hash. */
	switch_red_t		*redundancy_last;		/* pointer to the most recent Redundant Encoding packet in the hash. */
	uint32_t			redundancy_hash_sz;		/* Number of elements in the Redundant Encoding hash. */
	switch_inthash_t	*red_fec_seq_hash;		/* Hash of RED seq (hashed by network-ordered RED seq) of RED-FEC packets, from which FECs are stored in @redundancy_hash (hashed by FEc's seq). */
	switch_inthash_t	*save_for_fec_hash;		/* Store packets after they were played to be used to XOR incoming FEC packets */
	struct switch_jb_node_s	**rtp_mat_list;           /* Required packets to keep per RFC 5109 */
	uint16_t			rtp_mat_count;
	uint16_t 			rtp_mat_roll_index;
	switch_inthash_t	*missing_fecs_hash;		/* FEC seqs we didn't have when we looked for them, store them - they may be just late */
	uint32_t 			late_fec_count;			/* counter for late FEC packets */
	uint32_t			last_late_fec_count;
	uint32_t			avg_late_fec;			/* time diff between when we needed the FEC packet and when it arrived */
	uint32_t			fec_saved_the_day;		/* counter for number of times FEC recovery was effective */
	uint32_t			fec_didnt_save_the_day;	/* recovered but too late */
	uint32_t			redundancy_count;		/* total number of redundancy (FEC or FEC-only) rcvd. Used for stats */
	uint16_t			rtt_avg_ms;				/* RTT, will be used to decide if we should NACK packets being used for RTP material (earlier than JB decides they are missing.) */
	switch_bool_t		nack_fec_combo; 		/* enable NACK and FEC together */
#ifdef TEST_ULPFEC
	switch_inthash_t	*rtp_recovered_hash;	/* FEC-recovered packets, to be compared with NACK-recovered packets if we get them by NACK-ing. */ 
	uint32_t			rtp_recovered_count;	/* Number of times we inserted in the @rtp_recovered_hash. */
#endif // TEST_ULPFEC
};

#ifdef TEST_ULPFEC
static switch_status_t switch_jb_put_rtp_recovered(switch_jb_t *jb, switch_rtp_packet_t *rtp, switch_size_t rtp_len, uint8_t is_red);
static switch_status_t switch_jb_test_recovered(switch_jb_t *jb, switch_rtp_packet_t *rtp_nack, switch_rtp_packet_t *rtp_fec, switch_size_t rtp_nack_len, switch_size_t rtp_fec_len);
#endif

static void switch_jb_reset_redundancy_hash(switch_jb_t *jb);
static void switch_jb_free_redundancy_hash_entries(switch_jb_t *jb);

static int node_cmp(const void *l, const void *r)
{
	switch_jb_node_t *a = (switch_jb_node_t *) l;
	switch_jb_node_t *b = (switch_jb_node_t *) r;

	if (!a->visible) return 0;
	if (!b->visible) return 1;

	return ntohs(a->packet.header.seq) - ntohs(b->packet.header.seq);
}

//http://www.chiark.greenend.org.uk/~sgtatham/algorithms/listsort.c
switch_jb_node_t *sort_nodes(switch_jb_node_t *list, int (*cmp)(const void *, const void *)) {
	switch_jb_node_t *p, *q, *e, *tail;
	int insize, nmerges, psize, qsize, i;

	if (!list) {
		return NULL;
	}

	insize = 1;

	while (1) {
		p = list;
		list = NULL;
		tail = NULL;

		nmerges = 0;  /* count number of merges we do in this pass */

		while (p) {
			nmerges++;  /* there exists a merge to be done */
			            /* step `insize' places along from p */
			q = p;
			psize = 0;
			for (i = 0; i < insize; i++) {
				psize++;
				q = q->next;
				if (!q) break;
			}

			/* if q hasn't fallen off end, we have two lists to merge */
			qsize = insize;

			/* now we have two lists; merge them */
			while (psize > 0 || (qsize > 0 && q)) {

				/* decide whether next switch_jb_node_t of merge comes from p or q */
				if (psize == 0) {
					/* p is empty; e must come from q. */
					e = q; q = q->next; qsize--;
				} else if (qsize == 0 || !q) {
					/* q is empty; e must come from p. */
					e = p; p = p->next; psize--;
				} else if (cmp(p,q) <= 0) {
					/* First switch_jb_node_t of p is lower (or same);
					 * e must come from p. */
					e = p; p = p->next; psize--;
				} else {
					/* First switch_jb_node_t of q is lower; e must come from q. */
					e = q; q = q->next; qsize--;
				}

				/* add the next switch_jb_node_t to the merged list */
				if (tail) {
					tail->next = e;
				} else {
					list = e;
				}

				/* Maintain reverse pointers in a doubly linked list. */
				e->prev = tail;

				tail = e;
			}

			/* now p has stepped `insize' places along, and q has too */
			p = q;
		}

		tail->next = NULL;

		/* If we have done only one merge, we're finished. */
		if (nmerges <= 1)   /* allow for nmerges==0, the empty list case */
			return list;

		/* Otherwise repeat, merging lists twice the size */
		insize *= 2;
	}
}

// static inline void thin_frames(switch_jb_t *jb, int freq, int max);


static inline switch_jb_node_t *new_node(switch_jb_t *jb)
{
	switch_jb_node_t *np;

	switch_mutex_lock(jb->list_mutex);

	for (np = jb->node_list; np; np = np->next) {
		if (!np->visible) {
			break;
		}
	}

	if (!np) {
		int mult = 2;
		if (jb->type != SJB_VIDEO) {
			mult = 2;
		} else {
			if (jb->max_packet_len > mult) {
				mult = jb->max_packet_len;
			}
		}
		if (jb_has_fec(jb) && !(switch_test_flag(jb, SJB_QUEUE_ONLY))) {
			mult *= 2;
		} else if (switch_test_flag(jb, SJB_QUEUE_ONLY)) {
			mult = 2;
		}

		if (jb->allocated_nodes > jb->max_frame_len * mult) {
			jb_debug(jb, 2, "ALLOCATED FRAMES TOO HIGH! %d %d %d\n", jb->allocated_nodes, jb->max_frame_len, mult);
			if (jb_has_fec(jb) && (switch_test_flag(jb, SJB_QUEUE_ONLY))) {
				switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_WARNING, "ALLOCATED FRAMES TOO HIGH! RESET JB(vbw). %d %d %d\n", jb->allocated_nodes, jb->max_frame_len, mult);
			}
			switch_jb_reset(jb);
			switch_mutex_unlock(jb->list_mutex);
			return NULL;
		}
		
		np = switch_core_alloc(jb->pool, sizeof(*np));
		jb->allocated_nodes++;
		np->next = jb->node_list;
		if (np->next) {
			np->next->prev = np;
		}
		jb->node_list = np;

	}

	switch_assert(np);
	np->bad_hits = 0;
	np->visible = 1;
	jb->visible_nodes++;
	np->parent = jb;

	switch_mutex_unlock(jb->list_mutex);

	return np;
}

static inline void push_to_top(switch_jb_t *jb, switch_jb_node_t *node)
{
	if (node == jb->node_list) {
		jb->node_list = node->next;
	} else if (node->prev) {
		node->prev->next = node->next;
	}

	if (node->next) {
		node->next->prev = node->prev;
	}

	node->next = jb->node_list;
	node->prev = NULL;

	if (node->next) {
		node->next->prev = node;
	}

	jb->node_list = node;

	switch_assert(node->next != node);
	switch_assert(node->prev != node);
}

static inline void jb_hash_purge(switch_jb_t *jb, switch_inthash_t *hash, uint16_t start_seq, 
		uint16_t limit, const char *msg) 
{
	int i, more = 1;
	uint16_t remove_seq;

	switch_core_inthash_delete(hash, htons(ntohs(start_seq) - limit));
	while (switch_hashtable_count(hash) > limit * 2) {
		for (i = 0; i < limit; i++) {
			remove_seq = ntohs(start_seq) - i - limit * more;
			if (switch_core_inthash_delete(hash, htons(remove_seq))) {
				jb_debug(jb, 3, "%s: removed seq [%d]", msg, remove_seq);
			}
		}
		more++;
	}
}

static inline void hide_node(switch_jb_node_t *node, switch_bool_t pop)
{
	switch_jb_t *jb = node->parent;

	switch_mutex_lock(jb->list_mutex);

	if (node->visible) {
		node->visible = 0;
		node->bad_hits = 0;
		jb->visible_nodes--;

		if (pop) {
			push_to_top(jb, node);
		}
	}

	if (jb->node_hash_ts) {
		switch_core_inthash_delete(jb->node_hash_ts, node->packet.header.ts);
	}

	if (switch_core_inthash_delete(jb->node_hash, node->packet.header.seq)) {
		if (node->complete_frame_mark && jb->type == SJB_VIDEO) {
			jb->complete_frames--;
			node->complete_frame_mark = FALSE;
		}

		if (jb_has_fec(jb) && !switch_test_flag(jb, SJB_QUEUE_ONLY)) { 
			/* try to limit the FEC related hash sizes */
			/* and keep enough RTP "material" for recovery */ 
			jb_hash_purge(jb, jb->save_for_fec_hash, node->packet.header.seq, 200, "save_for_fec_hash");
			jb_hash_purge(jb, jb->red_fec_seq_hash, node->packet.header.seq, 200, "red_fec_seq_hash");
			/* not expected to be very big because we should adjust JB size */
			jb_hash_purge(jb, jb->missing_fecs_hash, node->packet.header.seq, 50, "missing_fecs_hash");
#ifdef TEST_ULPFEC
			jb_hash_purge(jb, jb->rtp_recovered_hash, node->packet.header.seq, 50, "rtp_recovered_hash");
#endif 
		}
	}

	switch_mutex_unlock(jb->list_mutex);
}

static inline void sort_free_nodes(switch_jb_t *jb)
{
	switch_mutex_lock(jb->list_mutex);
	jb->node_list = sort_nodes(jb->node_list, node_cmp);
	switch_mutex_unlock(jb->list_mutex);
}

static inline void hide_nodes(switch_jb_t *jb)
{
	switch_jb_node_t *np;

	switch_mutex_lock(jb->list_mutex);
	for (np = jb->node_list; np; np = np->next) {
		hide_node(np, SWITCH_FALSE);
	}
	switch_mutex_unlock(jb->list_mutex);
}

static inline void drop_ts(switch_jb_t *jb, uint32_t ts)
{
	switch_jb_node_t *np;
	int x = 0;

	switch_mutex_lock(jb->list_mutex);
	for (np = jb->node_list; np; np = np->next) {
		if (!np->visible) continue;

		if (ts == np->packet.header.ts) {
			hide_node(np, SWITCH_FALSE);
			x++;
		}
	}

	if (x) {
		sort_free_nodes(jb);
	}

	switch_mutex_unlock(jb->list_mutex);
}

static inline switch_jb_node_t *jb_find_lowest_seq(switch_jb_t *jb, uint32_t ts)
{
	switch_jb_node_t *np, *lowest = NULL;

	switch_mutex_lock(jb->list_mutex);
	for (np = jb->node_list; np; np = np->next) {
		if (!np->visible) continue;

		if (ts && ts != np->packet.header.ts) continue;

		if (!lowest || ntohs(lowest->packet.header.seq) > ntohs(np->packet.header.seq)) {
			lowest = np;
		}
	}
	switch_mutex_unlock(jb->list_mutex);

	return lowest;
}

static inline switch_jb_node_t *jb_find_lowest_node(switch_jb_t *jb)
{
	switch_jb_node_t *np, *lowest = NULL;

	switch_mutex_lock(jb->list_mutex);
	for (np = jb->node_list; np; np = np->next) {
		if (!np->visible) continue;

		if (!lowest || ntohl(lowest->packet.header.ts) > ntohl(np->packet.header.ts)) {
			lowest = np;
		}
	}
	switch_mutex_unlock(jb->list_mutex);

	return lowest ? lowest : NULL;
}

static inline uint32_t jb_find_lowest_ts(switch_jb_t *jb)
{
	switch_jb_node_t *lowest = jb_find_lowest_node(jb);

	return lowest ? lowest->packet.header.ts : 0;
}

#if 0
static inline void thin_frames(switch_jb_t *jb, int freq, int max)
{
	switch_jb_node_t *node, *this_node;
	int i = -1;
	int dropped = 0;

	switch_mutex_lock(jb->list_mutex);
	node = jb->node_list;

	while (node && dropped <= max) {
		this_node = node;
		node = node->next;
		
		if (this_node->visible) {
			i++;
		} else {
			continue;
		}

		if ((i % freq) == 0) {
			drop_ts(jb, this_node->packet.header.ts);
			dropped++;
		}
	}

	sort_free_nodes(jb);
	switch_mutex_unlock(jb->list_mutex);
}

static inline switch_jb_node_t *jb_find_highest_node(switch_jb_t *jb)
{
	switch_jb_node_t *np, *highest = NULL;

	switch_mutex_lock(jb->list_mutex);
	for (np = jb->node_list; np; np = np->next) {
		if (!np->visible) continue;

		if (!highest || ntohl(highest->packet.header.ts) < ntohl(np->packet.header.ts)) {
			highest = np;
		}
	}
	switch_mutex_unlock(jb->list_mutex);

	return highest ? highest : NULL;
}


static inline uint32_t jb_find_highest_ts(switch_jb_t *jb)
{
	switch_jb_node_t *highest = jb_find_highest_node(jb);

	return highest ? highest->packet.header.ts : 0;
}

static inline void drop_newest_frame(switch_jb_t *jb)
{
	uint32_t ts = jb_find_highest_ts(jb);

	drop_ts(jb, ts);
	jb_debug(jb, 1, "Dropping highest frame ts:%u\n", ntohl(ts));
}




static inline switch_jb_node_t *jb_find_penultimate_node(switch_jb_t *jb)
{
	switch_jb_node_t *np, *highest = NULL, *second_highest = NULL;

	switch_mutex_lock(jb->list_mutex);
	for (np = jb->node_list; np; np = np->next) {
		if (!np->visible) continue;

		if (!highest || ntohl(highest->packet.header.ts) < ntohl(np->packet.header.ts)) {
			if (highest) second_highest = highest;
			highest = np;
		}
	}
	switch_mutex_unlock(jb->list_mutex);

	return second_highest ? second_highest : highest;
}
#endif

static inline void jb_hit(switch_jb_t *jb)
{
	jb->period_good_count++;
	jb->consec_good_count++;
	jb->consec_miss_count = 0;
}

static void jb_frame_inc_line(switch_jb_t *jb, int i, int line)
{
	uint32_t old_frame_len = jb->frame_len;

	if (i == 0) {
		jb->frame_len = jb->min_frame_len;
		goto end;
	}

	if (i > 0) {
		if ((jb->frame_len + i) < jb->max_frame_len) {
			jb->frame_len += i;
		} else {
			jb->frame_len = jb->max_frame_len;
		}

		goto end;
	}

	if (i < 0) {
		if ((jb->frame_len + i) > jb->min_frame_len) {
			jb->frame_len += i;
		} else {
			jb->frame_len = jb->min_frame_len;
		}
	}

 end:

	if (jb->frame_len > jb->highest_frame_len) {
		jb->highest_frame_len = jb->frame_len;
	}

	if (old_frame_len != jb->frame_len) {
		jb_debug(jb, 1, "%d Change framelen from %u to %u\n", line, old_frame_len, jb->frame_len);

		//if (jb->session) {
		//	switch_core_session_request_video_refresh(jb->session);
		//}
	}

}

#define jb_frame_inc(_jb, _i) jb_frame_inc_line(_jb, _i, __LINE__)


static inline void jb_miss(switch_jb_t *jb)
{
	jb->period_miss_count++;
	jb->consec_miss_count++;
	jb->consec_good_count = 0;
}

#if 0
static inline int verify_oldest_frame(switch_jb_t *jb)
{
	switch_jb_node_t *lowest = NULL, *np = NULL;
	int r = 0;

	lowest = jb_find_lowest_node(jb);

	if (!lowest || !(lowest = jb_find_lowest_seq(jb, lowest->packet.header.ts))) {
		goto end;
	}

	switch_mutex_lock(jb->mutex);

	jb->node_list = sort_nodes(jb->node_list, node_cmp);

	for (np = lowest->next; np; np = np->next) {

		if (!np->visible) continue;

		if (ntohs(np->packet.header.seq) != ntohs(np->prev->packet.header.seq) + 1) {
			uint32_t val = (uint32_t)htons(ntohs(np->prev->packet.header.seq) + 1);

			if (!switch_core_inthash_find(jb->missing_seq_hash, val)) {
				switch_core_inthash_insert(jb->missing_seq_hash, val, (void *)(intptr_t)1);
			}
			break;
		}

		if (np->packet.header.ts != lowest->packet.header.ts || !np->next) {
			r = 1;
		}
	}

	switch_mutex_unlock(jb->mutex);

 end:

	return r;
}
#endif

static inline void drop_oldest_frame(switch_jb_t *jb)
{
	uint32_t ts = jb_find_lowest_ts(jb);

	drop_ts(jb, ts);
	jb_debug(jb, 1, "Dropping oldest frame ts:%u\n", ntohl(ts));
}



#if 0
static inline void drop_second_newest_frame(switch_jb_t *jb)
{
	switch_jb_node_t *second_newest = jb_find_penultimate_node(jb);

	if (second_newest) {
		drop_ts(jb, second_newest->packet.header.ts);
		jb_debug(jb, 1, "Dropping second highest frame ts:%u\n", ntohl(second_newest->packet.header.ts));
	}
}
#endif

static inline int check_seq(uint16_t a, uint16_t b)
{
	a = ntohs(a);
	b = ntohs(b);

	if (a >= b || (b > a && b > USHRT_MAX / 2 && a < USHRT_MAX / 2)) {
		return 1;
	}

	return 0;
}

static inline int check_ts(uint32_t a, uint32_t b)
{
	a = ntohl(a);
	b = ntohl(b);

	if (a > b || (b > a && b > UINT_MAX / 2 && a < UINT_MAX / 2)) {
		return 1;
	}

	return 0;
}

static inline void add_rtp_mat(switch_jb_t *jb, switch_rtp_packet_t *packet, switch_size_t len)
{
	switch_jb_node_t *node_pkt = NULL;
	if (jb->rtp_mat_count == RTP_MATERIAL_LIM) {
		/* reuse slot */
		jb->rtp_mat_roll_index = jb->rtp_mat_roll_index % (RTP_MATERIAL_LIM);
		node_pkt = jb->rtp_mat_list[jb->rtp_mat_roll_index];
		/* remove old packet before reusing this node */
		switch_core_inthash_delete(jb->save_for_fec_hash, node_pkt->packet.header.seq);
		jb->rtp_mat_roll_index++;
	} else {
		node_pkt = jb->rtp_mat_list[jb->rtp_mat_roll_index];
		jb->rtp_mat_count++;
		jb->rtp_mat_roll_index++;
	}
	node_pkt->packet = *packet;
	node_pkt->len = len;
	memcpy(node_pkt->packet.body, packet->body, len);
	switch_core_inthash_insert(jb->save_for_fec_hash, node_pkt->packet.header.seq, node_pkt);
}

static inline void add_node(switch_jb_t *jb, switch_rtp_packet_t *packet, switch_size_t len)
{
	switch_jb_node_t *node = new_node(jb);

	if (!node) {
		return;
	}


	node->packet = *packet;
	node->len = len;
	memcpy(node->packet.body, packet->body, len);

	switch_core_inthash_insert(jb->node_hash, node->packet.header.seq, node);
	if (jb->node_hash_ts) {
		switch_core_inthash_insert(jb->node_hash_ts, node->packet.header.ts, node);
	}

	jb_debug(jb, (packet->header.m ? 2 : 3), "PUT packet last_ts:%u ts:%u seq:%u%s\n",
			 ntohl(jb->highest_wrote_ts), ntohl(node->packet.header.ts), ntohs(node->packet.header.seq), packet->header.m ? " <MARK>" : "");

	if (jb->write_init && jb->type == SJB_VIDEO) {
		int seq_diff = 0, ts_diff = 0;

		if (ntohs(jb->highest_wrote_seq) > (USHRT_MAX - 100) && ntohs(packet->header.seq) < 100) {
			seq_diff = (USHRT_MAX - ntohs(jb->highest_wrote_seq)) + ntohs(packet->header.seq);
		} else {
			seq_diff = abs(((int)ntohs(packet->header.seq) - ntohs(jb->highest_wrote_seq)));
		}

		if (ntohl(jb->highest_wrote_ts) > (UINT_MAX - 1000) && ntohl(node->packet.header.ts) < 1000) {
			ts_diff = (UINT_MAX - ntohl(node->packet.header.ts)) + ntohl(node->packet.header.ts);
		} else {
			ts_diff = abs((int)((int64_t)ntohl(node->packet.header.ts) - (int64_t)ntohl(jb->highest_wrote_ts)));
		}

		if (((seq_diff >= 100) || (ts_diff > (900000 * 5)))) {
			jb_debug(jb, 2, "CHANGE DETECTED, PUNT %u\n", abs(((int)ntohs(packet->header.seq) - ntohs(jb->highest_wrote_seq))));
			switch_jb_reset(jb);
		}
	}

	if (!jb->write_init || ntohs(packet->header.seq) > ntohs(jb->highest_wrote_seq) ||
		(ntohs(jb->highest_wrote_seq) > USHRT_MAX - 100 && ntohs(packet->header.seq) < 100) ) {
		jb->highest_wrote_seq = packet->header.seq;
	}

	if (jb->type == SJB_VIDEO) {
		jb->packet_count++;
		
		if (jb->write_init && check_seq(packet->header.seq, jb->highest_wrote_seq) && check_ts(node->packet.header.ts, jb->highest_wrote_ts)) {
			jb_debug(jb, 2, "WRITE frame ts: %u complete=%u/%u n:%u\n", ntohl(node->packet.header.ts), jb->complete_frames , jb->frame_len, jb->visible_nodes);
			jb->highest_wrote_ts = packet->header.ts;
			jb->complete_frames++;

			jb->packet_count--;
			if (jb->packet_count > jb->max_packet_len) {
				jb->max_packet_len = jb->packet_count;
			}
			jb->packet_count = 1;
			node->complete_frame_mark = TRUE;
		} else if (!jb->write_init) {
			jb->highest_wrote_ts = packet->header.ts;
		}
	} else {
		if (jb->write_init || jb->type == SJB_TEXT || jb->type == SJB_AUDIO) {
			jb_debug(jb, 2, "WRITE frame ts: %u complete=%u/%u n:%u\n", ntohl(node->packet.header.ts), jb->complete_frames , jb->frame_len, jb->visible_nodes);
			jb->complete_frames++;
		} else {
			jb->highest_wrote_ts = packet->header.ts;
		}
	}

	if (!jb->write_init) jb->write_init = 1;
}

static inline void increment_ts(switch_jb_t *jb)
{
	if (!jb->target_ts) return;

	jb->last_psuedo_seq = jb->psuedo_seq;
	jb->last_target_ts = jb->target_ts;
	jb->target_ts = htonl((ntohl(jb->target_ts) + jb->samples_per_frame));
	jb->psuedo_seq++;
}

static inline void set_read_ts(switch_jb_t *jb, uint32_t ts)
{
	if (!ts) return;

	jb->last_psuedo_seq = jb->psuedo_seq;
	jb->last_target_ts = ts;
	jb->target_ts = htonl((ntohl(jb->last_target_ts) + jb->samples_per_frame));
	jb->psuedo_seq++;
}


static inline void increment_seq(switch_jb_t *jb)
{
	jb->last_target_seq = jb->target_seq;
	jb->target_seq = htons((ntohs(jb->target_seq) + 1));
}

static inline void set_read_seq(switch_jb_t *jb, uint16_t seq)
{
	jb->last_target_seq = seq;
	jb->target_seq = htons((ntohs(jb->last_target_seq) + 1));
}

static inline switch_status_t jb_next_packet_by_seq(switch_jb_t *jb, switch_jb_node_t **nodep)
{
	switch_jb_node_t *node = NULL;
	int sanity = 500;

 top:

	if (jb->type == SJB_VIDEO) {
		if (jb->dropped) {
			jb->dropped = 0;
			jb_debug(jb, 2, "%s", "DROPPED FRAME DETECTED RESYNCING\n");
			jb->target_seq = 0;

			if ((jb->session) && !(jb_has_fec(jb))) {
				switch_core_session_request_video_refresh(jb->session);
			}
		}
	}

	if (!jb->target_seq) {
		if ((node = switch_core_inthash_find(jb->node_hash, jb->target_seq))) {
			jb_debug(jb, 2, "FOUND rollover seq: %u\n", ntohs(jb->target_seq));
		} else if ((node = jb_find_lowest_seq(jb, 0))) {
			jb_debug(jb, 2, "No target seq using seq: %u as a starting point\n", ntohs(node->packet.header.seq));
		} else {
			jb_debug(jb, 1, "%s", "No nodes available....\n");
		}
		jb_hit(jb);
	} else if ((node = switch_core_inthash_find(jb->node_hash, jb->target_seq))) {
		jb_debug(jb, 2, "FOUND desired seq: %u\n", ntohs(jb->target_seq));
		jb_hit(jb);
	} else {

		if (jb_has_fec(jb) && switch_core_inthash_find(jb->red_fec_seq_hash, jb->target_seq)) {
			jb_debug(jb, 2, "SPURIOUS MISSING desired seq: %u\n", ntohs(jb->target_seq));
			increment_seq(jb);
			goto top;
		} else {
			jb_debug(jb, 2, "MISSING desired seq: %u\n", ntohs(jb->target_seq));
			jb_miss(jb);
		}

		if (jb->type == SJB_VIDEO) {
			int x;

			if ((jb->session) && !(jb_has_fec(jb))) {
				switch_core_session_request_video_refresh(jb->session);
			}

			for (x = 0; x < 10; x++) {
				
				increment_seq(jb);

				while (sanity && jb_has_fec(jb) && 
						switch_core_inthash_find(jb->red_fec_seq_hash, jb->target_seq)) {
					sanity--;
					jb_debug(jb, 2, "SKIP SEARCH incremental seq (FEC) %u\n", ntohs(jb->target_seq));
#if DEBUG_RED
					switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "JB RED: SKIP SEARCH incremental seq %u (FEC)\n", ntohs(jb->target_seq));
#endif
					increment_seq(jb);
				}

				if ((node = switch_core_inthash_find(jb->node_hash, jb->target_seq))) {
					jb_debug(jb, 2, "FOUND incremental seq: %u\n", ntohs(jb->target_seq));

					if (node->packet.header.m ||  node->packet.header.ts == jb->highest_read_ts) {
						jb_debug(jb, 2, "%s", "SAME FRAME DROPPING\n");
						jb->dropped++;
						drop_ts(jb, node->packet.header.ts);
						jb->highest_dropped_ts = ntohl(node->packet.header.ts);


						if (jb->period_miss_count > 2 && jb->period_miss_inc < 1) {
							jb->period_miss_inc++;
							jb_frame_inc(jb, 1);
						}

						node = NULL;
						goto top;
					}
					break;
				} else {
					jb_debug(jb, 2, "MISSING incremental seq: %u\n", ntohs(jb->target_seq));
				}
			}
			
		} else {
			increment_seq(jb);
		}
	}

	*nodep = node;

	if (node) {
		set_read_seq(jb, node->packet.header.seq);
		return SWITCH_STATUS_SUCCESS;
	}

	return SWITCH_STATUS_NOTFOUND;

}


static inline switch_status_t jb_next_packet_by_ts(switch_jb_t *jb, switch_jb_node_t **nodep)
{
	switch_jb_node_t *node = NULL;

	if (!jb->target_ts) {
		if ((node = jb_find_lowest_node(jb))) {
			jb_debug(jb, 2, "No target ts using ts: %u as a starting point\n", ntohl(node->packet.header.ts));
		} else {
			jb_debug(jb, 1, "%s", "No nodes available....\n");
		}
		jb_hit(jb);
	} else if ((node = switch_core_inthash_find(jb->node_hash_ts, jb->target_ts))) {
		jb_debug(jb, 2, "FOUND desired ts: %u\n", ntohl(jb->target_ts));
		jb_hit(jb);
	} else {
		jb_debug(jb, 2, "MISSING desired ts: %u\n", ntohl(jb->target_ts));
		jb_miss(jb);
		increment_ts(jb);
	}

	*nodep = node;

	if (node) {
		set_read_ts(jb, node->packet.header.ts);
		node->packet.header.seq = htons(jb->psuedo_seq);
		return SWITCH_STATUS_SUCCESS;
	}

	return SWITCH_STATUS_NOTFOUND;

}

static inline switch_status_t jb_next_packet(switch_jb_t *jb, switch_jb_node_t **nodep)
{
	if (jb->samples_per_frame) {
		return jb_next_packet_by_ts(jb, nodep);
	} else {
		return jb_next_packet_by_seq(jb, nodep);
	}
}

static inline void free_nodes(switch_jb_t *jb)
{
	switch_mutex_lock(jb->list_mutex);
	jb->node_list = NULL;
	switch_mutex_unlock(jb->list_mutex);
}

SWITCH_DECLARE(void) switch_jb_ts_mode(switch_jb_t *jb, uint32_t samples_per_frame, uint32_t samples_per_second)
{
	jb->samples_per_frame = samples_per_frame;
	jb->samples_per_second = samples_per_second;
	switch_core_inthash_init(&jb->node_hash_ts);
}

SWITCH_DECLARE(void) switch_jb_set_session(switch_jb_t *jb, switch_core_session_t *session)
{
	const char *var;

	if (session) {
		jb->session = session;
		jb->channel = switch_core_session_get_channel(session);
		if (jb->type == SJB_VIDEO && !switch_test_flag(jb, SJB_QUEUE_ONLY) && 
			(var = switch_channel_get_variable_dup(jb->channel, "jb_video_low_bitrate", SWITCH_FALSE, -1))) {
			int tmp = atoi(var);

			if (tmp >= 128 && tmp <= 10240) {
				jb->video_low_bitrate = (uint32_t)tmp;
			}
		}
	}
}

SWITCH_DECLARE(void) switch_jb_set_flag(switch_jb_t *jb, switch_jb_flag_t flag)
{
	switch_set_flag(jb, flag);
}

SWITCH_DECLARE(void) switch_jb_set_rtt(switch_jb_t *jb, uint16_t rtt)
{
	switch_mutex_lock(jb->mutex);
	jb->rtt_avg_ms = rtt * 1000;
	switch_mutex_unlock(jb->mutex);
}

SWITCH_DECLARE(void) switch_jb_clear_flag(switch_jb_t *jb, switch_jb_flag_t flag)
{
	switch_clear_flag(jb, flag);
}

SWITCH_DECLARE(int) switch_jb_poll(switch_jb_t *jb)
{
	if (jb->type == SJB_TEXT) {
		if (jb->complete_frames < jb->frame_len) {
			if (jb->complete_frames && !jb->buffer_lag) {
				jb->buffer_lag = 10;
			}
			if (jb->buffer_lag && --jb->buffer_lag == 0) {
				jb->flush = 1;
			}
		}
	}

	return (jb->complete_frames >= jb->frame_len) || jb->flush;
}

SWITCH_DECLARE(int) switch_jb_frame_count(switch_jb_t *jb)
{
	return jb->complete_frames;
}

SWITCH_DECLARE(void) switch_jb_debug_level(switch_jb_t *jb, uint8_t level)
{
	jb->debug_level = level;
}

SWITCH_DECLARE(void) switch_jb_reset(switch_jb_t *jb)
{

	if (jb->type == SJB_VIDEO) {
		switch_mutex_lock(jb->mutex);
		switch_core_inthash_destroy(&jb->missing_seq_hash);
		switch_core_inthash_init(&jb->missing_seq_hash);
		if (!switch_test_flag(jb, SJB_QUEUE_ONLY) && jb_has_fec(jb)) {
			switch_jb_reset_redundancy_hash(jb);
		}
		switch_mutex_unlock(jb->mutex);

		if (jb->session) {
			switch_core_session_request_video_refresh(jb->session);
		}
	}
	
	jb_debug(jb, 2, "%s", "RESET BUFFER\n");

	switch_mutex_lock(jb->mutex);
	hide_nodes(jb);
	switch_mutex_unlock(jb->mutex);
	
	jb->drop_flag = 0;
	jb->last_target_seq = 0;
	jb->target_seq = 0;
	jb->write_init = 0;
	jb->highest_wrote_seq = 0;
	jb->highest_wrote_ts = 0;
	jb->next_seq = 0;
	jb->highest_read_ts = 0;
	jb->highest_read_seq = 0;
	jb->read_init = 0;
	jb->next_seq = 0;
	jb->complete_frames = 0;
	jb->period_miss_count = 0;
	jb->consec_miss_count = 0;
	jb->period_miss_pct = 0;
	jb->period_good_count = 0;
	jb->consec_good_count = 0;
	jb->period_count = 0;
	jb->period_miss_inc = 0;
	jb->target_ts = 0;
	jb->last_target_ts = 0;

}

SWITCH_DECLARE(uint32_t) switch_jb_get_fec_rcvd_total(switch_jb_t *jb) 
{
	uint32_t fec_rcvd_total_count; /*counter of FEC, as we got them from RTP */
	switch_mutex_lock(jb->mutex);
	fec_rcvd_total_count = jb->redundancy_count;
	switch_mutex_unlock(jb->mutex);
	return fec_rcvd_total_count;
}

SWITCH_DECLARE(uint32_t) switch_jb_get_fec_success(switch_jb_t *jb) 
{
	uint32_t fec_recovered; /*count*/
	switch_mutex_lock(jb->mutex);
	fec_recovered = jb->fec_saved_the_day + jb->fec_didnt_save_the_day;
	switch_mutex_unlock(jb->mutex);
	return fec_recovered;
}

SWITCH_DECLARE(uint32_t) switch_jb_get_nack_success(switch_jb_t *jb) 
{
	uint32_t nack_recovered; /*count*/
	switch_mutex_lock(jb->mutex);
	nack_recovered = jb->nack_saved_the_day + jb->nack_didnt_save_the_day;
	switch_mutex_unlock(jb->mutex);
	return nack_recovered;
}

SWITCH_DECLARE(void) switch_jb_set_nack_fec_combo(switch_jb_t *jb) 
{
	switch_mutex_lock(jb->mutex);
	jb->nack_fec_combo = TRUE;
	switch_mutex_unlock(jb->mutex);
}

SWITCH_DECLARE(void) switch_jb_unset_nack_fec_combo(switch_jb_t *jb) 
{
	switch_mutex_lock(jb->mutex);
	jb->nack_fec_combo = FALSE;
	switch_mutex_unlock(jb->mutex);
}

SWITCH_DECLARE(uint32_t) switch_jb_get_packets_per_frame(switch_jb_t *jb) 
{
	uint32_t ppf;
	switch_mutex_lock(jb->mutex);
	ppf = jb->packet_count; /* get current packets per frame */
	switch_mutex_unlock(jb->mutex);
	return ppf;
}

SWITCH_DECLARE(switch_status_t) switch_jb_peek_frame(switch_jb_t *jb, uint32_t ts, uint16_t seq, int peek, switch_frame_t *frame)
{
	switch_jb_node_t *node = NULL;
	if (seq) {
		uint16_t want_seq = seq + peek;
		node = switch_core_inthash_find(jb->node_hash, htons(want_seq));
	} else if (ts && jb->samples_per_frame) {
		uint32_t want_ts = ts + (peek * jb->samples_per_frame);
		node = switch_core_inthash_find(jb->node_hash_ts, htonl(want_ts));
	}

	if (node) {
		frame->seq = ntohs(node->packet.header.seq);
		frame->timestamp = ntohl(node->packet.header.ts);
		frame->m = node->packet.header.m;
		frame->datalen = node->len - 12;

		if (frame->data && frame->buflen > node->len - 12) {
			memcpy(frame->data, node->packet.body, node->len - 12);
		}
		return SWITCH_STATUS_SUCCESS;
	}

	return SWITCH_STATUS_FALSE;
}

SWITCH_DECLARE(switch_status_t) switch_jb_get_frames(switch_jb_t *jb, uint32_t *min_frame_len, uint32_t *max_frame_len, uint32_t *cur_frame_len, uint32_t *highest_frame_len)
{

	switch_mutex_lock(jb->mutex);

	if (min_frame_len) {
		*min_frame_len = jb->min_frame_len;
	}

	if (max_frame_len) {
		*max_frame_len = jb->max_frame_len;
	}

	if (cur_frame_len) {
		*cur_frame_len = jb->frame_len;
	}

	switch_mutex_unlock(jb->mutex);

	return SWITCH_STATUS_SUCCESS;
}

SWITCH_DECLARE(switch_status_t) switch_jb_set_frames(switch_jb_t *jb, uint32_t min_frame_len, uint32_t max_frame_len)
{
	int lowest = 0;

	switch_mutex_lock(jb->mutex);

	if (jb->frame_len == jb->min_frame_len) lowest = 1;

	jb->min_frame_len = min_frame_len;
	jb->max_frame_len = max_frame_len;

	if (jb->frame_len > jb->max_frame_len) {
		jb->frame_len = jb->max_frame_len;
	}

	if (jb->frame_len < jb->min_frame_len) {
		jb->frame_len = jb->min_frame_len;
	}

	if (jb->frame_len > jb->highest_frame_len) {
		jb->highest_frame_len = jb->frame_len;
	}

	if (lowest) {
		jb->frame_len = jb->min_frame_len;
	}

	switch_mutex_unlock(jb->mutex);

	return SWITCH_STATUS_SUCCESS;
}

SWITCH_DECLARE(switch_status_t) switch_jb_create(switch_jb_t **jbp, switch_jb_type_t type,
												 uint32_t min_frame_len, uint32_t max_frame_len, switch_memory_pool_t *pool,
												 ulp_policy_t *ulp_policy)
{
	switch_jb_t *jb;
	int free_pool = 0;

	if (!pool) {
		switch_core_new_memory_pool(&pool);
		free_pool = 1;
	}

	jb = switch_core_alloc(pool, sizeof(*jb));
	jb->free_pool = free_pool;
	jb->min_frame_len = jb->frame_len = min_frame_len;
	jb->max_frame_len = max_frame_len;
	jb->pool = pool;
	jb->type = type;
	jb->highest_frame_len = jb->frame_len;

	if (jb->type == SJB_VIDEO) {
		switch_core_inthash_init(&jb->missing_seq_hash);
		jb->period_len = 2500;
	} else {
		jb->period_len = 250;
	}
	
	switch_core_inthash_init(&jb->node_hash);
	switch_mutex_init(&jb->mutex, SWITCH_MUTEX_NESTED, pool);
	switch_mutex_init(&jb->list_mutex, SWITCH_MUTEX_NESTED, pool);

	if (jb->type == SJB_VIDEO && ulp_policy) {
		int i;
		switch_core_inthash_init(&jb->redundancy_hash);
		jb->redundancy_hash_sz = 0;
		jb->redundancy_first = NULL;
		jb->redundancy_last = NULL;
		switch_core_inthash_init(&jb->save_for_fec_hash);
		switch_core_inthash_init(&jb->missing_fecs_hash);
		jb->ulp_policy = ulp_policy;
		ulp_set_user_data(ulp_policy, jb);
		switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_DEBUG1, "JB RED: associating ULP policy %p and jb=%p\n", (void *) ulp_policy, (void*) jb);
		switch_core_inthash_init(&jb->red_fec_seq_hash);
		jb->rtp_mat_list = switch_core_alloc(jb->pool, RTP_MATERIAL_LIM * sizeof(void*)); 
		for (i = 0; i < RTP_MATERIAL_LIM; i++) {
			jb->rtp_mat_list[i] = switch_core_alloc(jb->pool, sizeof(switch_jb_node_t));
		}
		switch_jb_set_nack_fec_combo(jb);
		switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_INFO, "JB RED: ULPFEC ready.\n");
#ifdef TEST_ULPFEC
		switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_WARNING, 
			"JB RED: built with TEST_ULPFEC, will not add recovered packets to JB\n");
		switch_core_inthash_init(&jb->rtp_recovered_hash);
		jb->rtp_recovered_count = 0;
#endif
	}
	*jbp = jb;

	return SWITCH_STATUS_SUCCESS;
}

SWITCH_DECLARE(switch_status_t) switch_jb_destroy(switch_jb_t **jbp)
{
	switch_jb_t *jb = *jbp;
	*jbp = NULL;

	if (jb->type == SJB_VIDEO && !switch_test_flag(jb, SJB_QUEUE_ONLY) && jb_has_fec(jb)) {
		jb_debug(jb, 3, "Stats: NACK saved the day: %u\n", jb->nack_saved_the_day);
		jb_debug(jb, 3, "Stats: NACK was late: %u\n", jb->nack_didnt_save_the_day);
		jb_debug(jb, 3, "Stats: FEC saved the day: %u\n", jb->fec_saved_the_day);
		jb_debug(jb, 3, "Stats: FEC-recovered was late: %u\n", jb->fec_didnt_save_the_day);
		jb_debug(jb, 3, "Stats: Hash entrycount: node_hash %u\n", switch_hashtable_count(jb->node_hash));
		jb_debug(jb, 3, "Stats: Hash entrycount: missing_fecs_hash %u\n", switch_hashtable_count(jb->missing_fecs_hash));
		jb_debug(jb, 3, "Stats: Hash entrycount: missing_seq_hash %u\n", switch_hashtable_count(jb->missing_seq_hash));
		jb_debug(jb, 3, "Stats: Hash entrycount: red_fec_seq_hash %u\n", switch_hashtable_count(jb->red_fec_seq_hash));
		jb_debug(jb, 3, "Stats: Hash entrycount: redundancy_hash %u\n", switch_hashtable_count(jb->redundancy_hash));
		jb_debug(jb, 3, "Stats: Hash entrycount: save_for_fec_hash %u\n", switch_hashtable_count(jb->save_for_fec_hash));
	}

	if (jb->type == SJB_VIDEO) {
		switch_core_inthash_destroy(&jb->missing_seq_hash);
	}
	switch_core_inthash_destroy(&jb->node_hash);

	if (jb->node_hash_ts) {
		switch_core_inthash_destroy(&jb->node_hash_ts);
	}

	if (jb_has_fec(jb)) {
		if (jb->redundancy_hash) {
			switch_jb_free_redundancy_hash_entries(jb);
			switch_core_inthash_destroy(&jb->redundancy_hash);
		}
		jb->redundancy_hash_sz = 0;
		jb->redundancy_first = NULL;
		jb->redundancy_last = NULL;
		if (jb->red_fec_seq_hash) {
			switch_core_inthash_destroy(&jb->red_fec_seq_hash);
		}
		if (jb->save_for_fec_hash) {
			switch_core_inthash_destroy(&jb->save_for_fec_hash);
		}
		if (jb->missing_fecs_hash) {
			if (jb->late_fec_count) {
				jb_debug(jb, 3, "Unsuccessful attempts due to late FEC: %u\n", jb->late_fec_count);
			}
			switch_core_inthash_destroy(&jb->missing_fecs_hash);
		}
#ifdef TEST_ULPFEC
		if (jb->rtp_recovered_hash) {
			switch_core_inthash_destroy(&jb->rtp_recovered_hash);
			jb_debug(jb, 3, "FEC-recovered packet count: %u\n", jb->rtp_recovered_count);
		}
		jb->rtp_recovered_count = 0;
#endif
	}
	free_nodes(jb);

	if (jb->free_pool) {
		switch_core_destroy_memory_pool(&jb->pool);
	}

	return SWITCH_STATUS_SUCCESS;
}

SWITCH_DECLARE(uint32_t) switch_jb_pop_nack(switch_jb_t *jb)
{
	switch_hash_index_t *hi = NULL;
	uint32_t nack = 0;
	uint16_t blp = 0;
	uint16_t least = 0;
	int i = 0;
	void *val;
	const void *var;
	uint16_t wait = jb->frame_len;

	if (jb->type != SJB_VIDEO) {
		return 0;
	}

	switch_mutex_lock(jb->mutex);

#ifdef TEST_ULPFEC
	wait = jb->max_frame_len * 2;
#else
	if (jb_has_fec(jb) && jb->nack_fec_combo) {
		wait = jb->max_frame_len * 2;
	} else if (jb_has_fec(jb)) {
		switch_mutex_unlock(jb->mutex);
		return 0;
	}
#endif 

 top:

	for (hi = switch_core_hash_first_iter(jb->missing_seq_hash, hi); hi; hi = switch_core_hash_next(&hi)) {
		uint16_t seq;
		//const char *token;
		switch_time_t then = 0;

		switch_core_hash_this(hi, &var, NULL, &val);
		//token = (const char *) val;

		//if (token == TOKEN_2) {
			//switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "SKIP %u %s\n", ntohs(*((uint16_t *) var)), token);
			//printf("WTf\n");
		//	continue;
		//}

		seq = ntohs(*((uint16_t *) var));
		then = (intptr_t) val;

		if (then != 1) {
			if (((uint32_t)(switch_time_now() - then)) < RENACK_TIME) {
				jb_debug(jb, 3, "NACKABLE seq %u too soon to repeat\n", seq);
				continue;
			}
//			jb_debug(jb, 3, "NACKABLE seq %u not too soon to repeat %u\n", seq, (uint32_t)(switch_time_now() - then));
#ifdef TEST_ULPFEC
			/*expire later: more nack for the test */
			if (((uint32_t)(switch_time_now() - then)) > (RENACK_TIME * 30)) {
				switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(seq));
				goto top;
			}
#else
			if (jb_has_fec(jb) && jb->rtt_avg_ms && ((uint32_t)(switch_time_now() - then) >= jb->rtt_avg_ms * 1000)) {
				switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(seq));
				goto top;
			} else if (((uint32_t)(switch_time_now() - then)) > (RENACK_TIME * 10)) {
				switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(seq));
				goto top;
			}
#endif
		}

		if (seq < ntohs(jb->target_seq) - wait) { 
			/* expire later, so that FEC code checking missing_seq_hash recursively will try recovery */
			jb_debug(jb, 3, "NACKABLE seq %u expired\n", seq);
			switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(seq));
			goto top;
		}

		if (!least || seq < least) {
			least = seq;
		}
	}

	switch_safe_free(hi);

	if (least && switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(least))) {
		jb_debug(jb, 3, "Found NACKABLE seq %u\n", least);
		nack = (uint32_t) htons(least);
		switch_core_inthash_insert(jb->missing_seq_hash, nack, (void *) (intptr_t)switch_time_now());

		for(i = 0; i < 16; i++) {
			if (switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(least + i + 1))) {
				switch_core_inthash_insert(jb->missing_seq_hash, (uint32_t)htons(least + i + 1), (void *)(intptr_t)switch_time_now());
				jb_debug(jb, 3, "Found addtl NACKABLE seq %u\n", least + i + 1);
				blp |= (1 << i);
			}
		}

		blp = htons(blp);
		nack |= (uint32_t) blp << 16;

		//jb_frame_inc(jb, 1);
	}

	switch_mutex_unlock(jb->mutex);


	return nack;
}

SWITCH_DECLARE(switch_status_t) switch_jb_put_packet(switch_jb_t *jb, switch_rtp_packet_t *packet, switch_size_t len, switch_bool_t fec)
{
	uint32_t i;
	uint16_t want = ntohs(jb->next_seq), got = ntohs(packet->header.seq);
	uint32_t fec_n = 0;	/* Number of FECs between @want and @got - these packets are in @redundancy_hash, not in jb. */

#ifdef TEST_ULPFEC
	switch_rtp_for_fec_t *e = NULL;
	switch_rtp_packet_t *rtp_fec = NULL;	/* RTP recovered by FEC. */
	switch_rtp_packet_t *rtp_nack = NULL;	/* RTP recovered by NACK. */
	switch_size_t		rtp_fec_len = 0;	/* Payload length. */
	switch_size_t		rtp_nack_len = 0;	/* Payload length. */
#endif // TEST_ULPFEC

	if (len >= sizeof(switch_rtp_packet_t)) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "trying to put %" SWITCH_SIZE_T_FMT " bytes exceeding buffer, truncate to %" SWITCH_SIZE_T_FMT "\n", len, sizeof(switch_rtp_packet_t));
		len = sizeof(switch_rtp_packet_t);
	}

	switch_mutex_lock(jb->mutex);

#ifdef TEST_ULPFEC

	if (jb->type == SJB_VIDEO && jb->rtp_recovered_hash) {
	
		//switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "VP8 seq=%u	ts=%u\n", ntohs(packet->header.seq), ntohl(packet->header.ts));

		rtp_nack = packet;
		rtp_nack_len = len;

		if ((e = switch_core_inthash_find(jb->rtp_recovered_hash, packet->header.seq))) {
		
			rtp_fec = &e->rtp;
			rtp_fec_len = e->rtp_len + 12;
			switch_jb_test_recovered(jb, rtp_nack, rtp_fec, rtp_nack_len, rtp_fec_len);
		}
	}

#endif // TEST_ULPFEC

	if (jb->highest_dropped_ts) {
		if (ntohl(packet->header.ts) < jb->highest_dropped_ts) {
			jb_debug(jb, 2, "%s", "TS ALREADY DROPPED, DROPPING PACKET\n");
			switch_mutex_unlock(jb->mutex);
			return SWITCH_STATUS_SUCCESS;
		}
		jb->highest_dropped_ts = 0;
	}


	if (!want) want = got;

	if (switch_test_flag(jb, SJB_QUEUE_ONLY) || jb->type == SJB_AUDIO || jb->type == SJB_TEXT) {
		jb->next_seq = htons(got + 1);
	} else {

				
		if (!fec && switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(got))) {
			if (got < ntohs(jb->target_seq)) {
				jb_debug(jb, 2, "got nacked seq %u too late\n", got);
				jb_frame_inc(jb, 1);
				jb->nack_didnt_save_the_day++;
			} else {
				jb_debug(jb, 2, "got nacked seq %u saved the day!\n", got);
				jb->nack_saved_the_day++;
			}
		} else if (fec) {
			/* recovered FEC which was not marked as missed yet */
			switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t)htons(got));
			if (got < ntohs(jb->target_seq)) {
				jb_debug(jb, 2, "got fec-recovered seq %u too late\n", got);
				jb_frame_inc(jb, 1);
				jb->fec_didnt_save_the_day++;
			} else {
				jb_debug(jb, 2, "got fec-recovered seq %u saved the day!\n", got);
				jb->fec_saved_the_day++;
			}
		}

		if (got > want) {
				
			if (jb_has_fec(jb)) {
				for (i = want; i < got; i++) {
					if (switch_core_inthash_find(jb->red_fec_seq_hash, (uint32_t) htons(i))) {
						fec_n++;
					}
				}
			}
			if (got - want > jb->max_frame_len + fec_n && got - want > RTP_MATERIAL_LIM) {
				jb_debug(jb, 2, "Missing %u frames, Resetting\n", got - want - fec_n);
				switch_jb_reset(jb);
			} else {
				if (jb->type == SJB_VIDEO && jb->frame_len + fec_n < got - want) {
					jb_frame_inc(jb, 1);
				}
				jb_debug(jb, 2, "GOT %u WANTED %u; MARK SEQS MISSING %u - %u\n", got, want, want, got - 1);

				for (i = want; i < got; i++) {
					if (jb_has_fec(jb) && 
							switch_core_inthash_find(jb->red_fec_seq_hash, (uint32_t) htons(i))) {
						
						jb_debug(jb, 2, "UNMARK MISSING (FEC) %u\n", i);
#if DEBUG_RED
						switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "JB RED: UNMARK MISSING %u (FEC)\n", i);
#endif
						switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t) htons(i));

					} else {

						jb_debug(jb, 2, "MARK MISSING %u ts:%u\n", i, ntohl(packet->header.ts));
						switch_core_inthash_insert(jb->missing_seq_hash, (uint32_t)htons(i), (void *)(intptr_t)1);
					}
				}
			}
		}

		if (got >= want || (want - got) > 1000 + fec_n) {
			jb->next_seq = htons(got + 1);
		}
	}

	if (jb_has_fec(jb) && !switch_test_flag(jb, SJB_QUEUE_ONLY) && len > 12) { 
		add_rtp_mat(jb, packet, len);
	}

	add_node(jb, packet, len);

	/* accomodate a more variable packet len for FEC. A frame can be bigger with several packets vs non-FEC.
	 * Make sure we make enough room for the next frame (which may be bigger than the one we just dropped)
	 * and avoid to reset JB (vbw). The way we're dropping frames from VBW should work for non-FEC too. */
	if (switch_test_flag(jb, SJB_QUEUE_ONLY) && jb->max_packet_len && jb->max_frame_len * 2 > jb->max_packet_len &&
			jb->allocated_nodes > jb->max_frame_len * 2 - 1) {
		while ((jb->max_frame_len * 2 - jb->visible_nodes) < jb->max_packet_len) {
			drop_oldest_frame(jb);
		}
	} else if (switch_test_flag(jb, SJB_QUEUE_ONLY) && jb->max_packet_len && jb->max_frame_len * 2 < jb->max_packet_len) {
		/* rtp_nack_buffer_size less than initial max_packet_len */
		drop_oldest_frame(jb);
	}

	switch_mutex_unlock(jb->mutex);

	return SWITCH_STATUS_SUCCESS;
}

SWITCH_DECLARE(switch_status_t) switch_jb_get_packet_by_seq(switch_jb_t *jb, uint16_t seq, switch_rtp_packet_t *packet, switch_size_t *len)
{
	switch_jb_node_t *node;
	switch_status_t status = SWITCH_STATUS_NOTFOUND;

	switch_mutex_lock(jb->mutex);
	if ((node = switch_core_inthash_find(jb->node_hash, seq))) {
		jb_debug(jb, 2, "Found buffered seq: %u\n", ntohs(seq));
		*packet = node->packet;
		*len = node->len;
		memcpy(packet->body, node->packet.body, node->len);
		packet->header.version = 2;
		status = SWITCH_STATUS_SUCCESS;
	} else {
		jb_debug(jb, 2, "Missing buffered seq: %u\n", ntohs(seq));
	}
	switch_mutex_unlock(jb->mutex);

	return status;
}

SWITCH_DECLARE(switch_size_t) switch_jb_get_last_read_len(switch_jb_t *jb)
{
	return jb->last_len;
}

SWITCH_DECLARE(switch_status_t) switch_jb_get_packet(switch_jb_t *jb, switch_rtp_packet_t *packet, switch_size_t *len)
{
	switch_jb_node_t *node = NULL;
	switch_status_t status;
	int plc = 0;
	
	switch_mutex_lock(jb->mutex);

	if (jb->complete_frames == 0) {
		jb->flush = 0;
		switch_goto_status(SWITCH_STATUS_BREAK, end);
	}

	if (jb->complete_frames < jb->frame_len) {

		switch_jb_poll(jb);

		if (!jb->flush) {
			jb_debug(jb, 2, "BUFFERING %u/%u\n", jb->complete_frames , jb->frame_len);
			switch_goto_status(SWITCH_STATUS_MORE_DATA, end);
		}
	}

	jb_debug(jb, 2, "GET PACKET %u/%u n:%d\n", jb->complete_frames , jb->frame_len, jb->visible_nodes);

	if (++jb->period_count >= jb->period_len) {

		if (jb->consec_good_count >= (jb->period_len - 5)) {
			jb_frame_inc(jb, -1);
		}

		jb->period_count = 1;
		jb->period_miss_inc = 0;
		jb->period_miss_count = 0;
		jb->period_good_count = 0;
		jb->consec_miss_count = 0;
		jb->consec_good_count = 0;

		if (jb->type == SJB_VIDEO && jb->channel && jb->video_low_bitrate) {
			//switch_time_t now = switch_time_now();
			//int ok = (now - jb->last_bitrate_change) > 10000;

			if (switch_channel_test_flag(jb->channel, CF_VIDEO_BITRATE_UNMANAGABLE) && jb->frame_len == jb->min_frame_len) {
				jb_debug(jb, 2, "%s", "Allow BITRATE changes\n");
				switch_channel_clear_flag_recursive(jb->channel, CF_VIDEO_BITRATE_UNMANAGABLE);
				jb->bitrate_control = 0;
				if (jb->session) {
					switch_core_session_request_video_refresh(jb->session);
				}
			} else if (!switch_channel_test_flag(jb->channel, CF_VIDEO_BITRATE_UNMANAGABLE) && jb->frame_len > jb->max_frame_len / 2) {
				switch_core_session_message_t msg = { 0 };

				jb->bitrate_control = jb->video_low_bitrate;
				
				msg.message_id = SWITCH_MESSAGE_INDICATE_BITRATE_REQ;
				msg.numeric_arg = jb->bitrate_control * 1024;
				msg.from = __FILE__;

				jb_debug(jb, 2, "Force BITRATE to %d\n", jb->bitrate_control);

				switch_core_session_receive_message(jb->session, &msg);
				switch_channel_set_flag_recursive(jb->channel, CF_VIDEO_BITRATE_UNMANAGABLE);
				if (jb->session) {
					switch_core_session_request_video_refresh(jb->session);
				}
			}
		}
	}


	jb->period_miss_pct = ((double)jb->period_miss_count / jb->period_count) * 100;

	//if (jb->period_miss_pct > 60.0f) {
	//	jb_debug(jb, 2, "Miss percent %02f too high, resetting buffer.\n", jb->period_miss_pct);
	//	switch_jb_reset(jb);
	//}

	if ((status = jb_next_packet(jb, &node)) == SWITCH_STATUS_SUCCESS) {
		jb_debug(jb, 2, "Found next frame cur ts: %u seq: %u\n", htonl(node->packet.header.ts), htons(node->packet.header.seq));

		if (!jb->read_init || check_seq(node->packet.header.seq, jb->highest_read_seq)) {
			jb->highest_read_seq = node->packet.header.seq;
		}

		if (jb->type != SJB_VIDEO ||
			(jb->read_init && check_seq(node->packet.header.seq, jb->highest_read_seq) && check_ts(node->packet.header.ts, jb->highest_read_ts))) {

			if (jb->type != SJB_VIDEO) {
				jb->complete_frames--;
			}
			jb_debug(jb, 2, "READ frame ts: %u complete=%u/%u n:%u\n", ntohl(node->packet.header.ts), jb->complete_frames , jb->frame_len, jb->visible_nodes);
			jb->highest_read_ts = node->packet.header.ts;
		} else if (!jb->read_init) {
			jb->highest_read_ts = node->packet.header.ts;
		}

		if (!jb->read_init) jb->read_init = 1;
	} else {
		if (jb->type == SJB_VIDEO) {
			//switch_jb_reset(jb);

			switch(status) {
			case SWITCH_STATUS_RESTART:
				jb_debug(jb, 2, "%s", "Error encountered ask for new keyframe\n");
				switch_goto_status(SWITCH_STATUS_RESTART, end);
			case SWITCH_STATUS_NOTFOUND:
			default:
				jb_debug(jb, 2, "%s", "No frames found wait for more\n");
				switch_goto_status(SWITCH_STATUS_MORE_DATA, end);
			}
		} else {
			switch(status) {
			case SWITCH_STATUS_RESTART:
				jb_debug(jb, 2, "%s", "Error encountered\n");
				switch_jb_reset(jb);
				switch_goto_status(SWITCH_STATUS_RESTART, end);
			case SWITCH_STATUS_NOTFOUND:
			default:
				if (jb->consec_miss_count > jb->frame_len) {
					//switch_jb_reset(jb);
					jb_frame_inc(jb, 1);
					jb_debug(jb, 2, "%s", "Too many frames not found, RESIZE\n");
					switch_goto_status(SWITCH_STATUS_RESTART, end);
				} else {
					jb_debug(jb, 2, "%s", "Frame not found suggest PLC\n");
					plc = 1;
					switch_goto_status(SWITCH_STATUS_NOTFOUND, end);
				}
			}
		}
	}

	if (node) {
		status = SWITCH_STATUS_SUCCESS;

		*packet = node->packet;
		*len = node->len;
		jb->last_len = *len;
		memcpy(packet->body, node->packet.body, node->len);
		packet->header.version = 2;
		hide_node(node, SWITCH_TRUE);

		jb_debug(jb, 2, "GET packet ts:%u seq:%u %s\n", ntohl(packet->header.ts), ntohs(packet->header.seq), packet->header.m ? " <MARK>" : "");

	} else {
		status = SWITCH_STATUS_MORE_DATA;
	}

 end:

	if (plc) {
		uint16_t seq;
		uint32_t ts = 0;

		if (jb->samples_per_frame) {
			seq = htons(jb->last_psuedo_seq);
			ts = jb->last_target_ts;
		} else {
			seq = jb->last_target_seq;
		}

		packet->header.seq = seq;
		packet->header.ts = ts;
	}

	switch_mutex_unlock(jb->mutex);
	
	if (jb->type == SJB_VIDEO) {
		if (jb_has_fec(jb) && jb->late_fec_count > jb->last_late_fec_count) {
			jb->last_late_fec_count = jb->late_fec_count;
			jb_frame_inc(jb, 1); 
		}
		if (jb->complete_frames > jb->max_frame_len * 2) {
			jb_debug(jb, 2, "JB TOO BIG (%d), RESET\n", jb->complete_frames);
			switch_jb_reset(jb);
		}
	} else {
		int too_big = (int)(jb->max_frame_len * 1.5);
		if (jb->visible_nodes > too_big && status == SWITCH_STATUS_SUCCESS) {
			status = SWITCH_STATUS_TIMEOUT;
		}
	}
	
	return status;
}

/* ULP Callbacks. */

/**
 * Return copies of RTP packets needed by FEC with @snb on protection level @k. ULP owns them and calls free on them.
 * @policy - ULP context describing the levels of protection, storing pointer to jb, etc.
 * @snb - the RTP seq in the header of the FEC packet, indicates the lowest RTP sequence used by this FEC packet
 * @k - level of protection the FEC is looked for
 * @rtp - pointer to existing RTP array passed from ULP, this callback must fill this array with entries and return number of entries inserted
 * @miss_n - allowed misses on searching of RTP packets (if this is done for recovery,
 * then we need one packet less size of the group, so we allow for one miss. */
static uint8_t jb_ulp_rtp_material(ulp_policy_t *policy, uint16_t snb, uint8_t k, ulp_rtp_msg_t *rtp, int miss_n)
{
	uint8_t         rtp_n = 0;                          /* number of RTP packets covering @seq and present in the RTP buffer */
	uint8_t         gk = 0;								/* number of RTP packets in a group */
	uint16_t		seq = 0;

	switch_rtp_packet_t packet;
	switch_size_t len = 0;
	int i = 0;
	int missed = 0;
	uint16_t mask16 = policy->lvl[k].mask16;
	uint32_t mask_cont = policy->lvl[k].mask_cont;
	switch_jb_node_t *node;
	
	switch_jb_t *jb = ulp_get_user_data(policy);
	if (!jb) {
		return 0;
	}


	memset(&packet, 0, sizeof(packet));
	gk = policy->lvl[k].gk;								/* Get the number of RTP packets needed by this policy on @k level of protection. */

	/* If this is called in result of FEC generation, then this method
	 * must return @gk number of RTP packets in order for the ulp_gen_fec() call
	 * to successfully generate FEC packet.
	 * If it's called during packet recovery, @gk - 1 packets should be returned.
	 */

	rtp_n = 0;

	/* We need @i variable here to walk through complete sequence of seq in range [@snb + G(k_max) - G(k), @snb + G(k_max)).
	 * @rtp_n indicates the number of packets found. */ 
	i = 0;
	while ((i < gk) && (missed <= miss_n)) {
	
		seq = ulp_mask_2_seq(mask16, mask_cont, snb, i + 1);

		if ((node = switch_core_inthash_find(jb->save_for_fec_hash, htons(seq)))) {
			packet = node->packet;
			if (node->len <= 12) {
				// invalid, this should not happen, we only put packets with len > 12 in this hash
				++missed;
				++i;
				continue;
			}
			len = node->len - 12;
			memcpy(packet.body, node->packet.body, node->len);
			packet.header.version = 2;
			jb_debug(jb, 2, "Found buffered seq in fec save hash: seq:%u len:%u\n", seq, (unsigned int)len);
		} else {
			++missed;
			++i;
			if (jb->nack_fec_combo) {
				/* short-circuit the missing_seq_hash to request RTP material via NACK (need for early recovery) */
				if (switch_core_inthash_find(jb->missing_seq_hash, (uint32_t) htons(seq))) {
					switch_core_inthash_insert(jb->missing_seq_hash, htons(seq), (void *) (intptr_t)switch_time_now());
				}
			}
			continue;
		}
		rtp[rtp_n].rtp = malloc(sizeof(ulp_rtp_hdr_t) + len);    
		if (!rtp[rtp_n].rtp) {
			goto jb_rtp_m_err;
		}
		memcpy(&rtp[rtp_n].rtp->hdr, &packet.header, 12);
		memcpy(rtp[rtp_n].rtp->b, packet.body, len);

		rtp[rtp_n].payload_len = rtp[rtp_n].payload_original_len = len;
		++rtp_n;
		++i;
	}

	return rtp_n;

jb_rtp_m_err:

	while (rtp_n) {
		free(rtp[rtp_n - 1].rtp);
		memset(&rtp[rtp_n - 1], 0, sizeof(rtp[0]));
		rtp_n--;
	}
	return 0;
}


/* ULP receiver's callbacks. */

SWITCH_DECLARE(ulp_rtp_pkt_t*) jb_ulp_cb_rtp_by_seq(ulp_policy_t *policy, uint16_t seq, uint16_t *payload_original_len)
{
    ulp_rtp_pkt_t   *rtp = NULL;

	switch_rtp_packet_t packet;
	switch_size_t len = 0;


	switch_jb_t *jb = ulp_get_user_data(policy);
	if (!jb) {
		return NULL;
	}

	memset(&packet, 0, sizeof(packet));
	if (switch_jb_get_packet_by_seq(jb, htons(seq), &packet, &len) != SWITCH_STATUS_SUCCESS) {
		return NULL;
	}

	if (len <= 12) {
		return NULL;
	}

	rtp = malloc(sizeof(ulp_rtp_hdr_t) + len);
	if (!rtp) {
		return NULL;
	}
	memcpy(&rtp->hdr, &packet.header, 12);
	memcpy(rtp->b, packet.body, len);
	*payload_original_len = len;
	
	return rtp;
}

SWITCH_DECLARE(uint8_t) jb_ulp_cb_find_fecs_for_seq_on_lvl(ulp_policy_t *policy, uint16_t wanted_seq, uint8_t k, ulp_rtp_msg_t *fecs)
{
	switch_red_t	*e = NULL;
	ulp_fec_pkt_t *fec = NULL;
	uint8_t fec_n = 0;
	uint16_t fec_payload_len = 0;
	uint16_t roll_snb = (wanted_seq > 46 ? wanted_seq - 47 : 0);


	switch_jb_t *jb = ulp_get_user_data(policy);
	if (!jb) {
		return 0;
	}
	if (!jb->redundancy_hash) {
		return 0;
	}

	/**
	 * Search for all FECs in hashtable which protect @seq on level @k.
	 * This method should return all FEC packets which protect @seq on level @k.
	 */
	for (; roll_snb < wanted_seq + 1; ++roll_snb) {

		switch_mutex_lock(jb->mutex);
		e = switch_core_inthash_find(jb->redundancy_hash, (uint32_t) htons(roll_snb));

		for (; e; e = e->next) {
			fec = (ulp_fec_pkt_t *) e->redundancy;
			fec_payload_len = e->redundancy_len - sizeof(ulp_fec_hdr_t);

			if (fec && ulp_seq_protected_by_fec_on_lvl(policy, fec, fec_payload_len, wanted_seq, k)) {
				fecs[fec_n].rtp = malloc(sizeof(ulp_fec_hdr_t) + fec_payload_len);				
				if (!fecs[fec_n].rtp) {
					goto jb_ulp_cb_find_fec_err;
				}
				memcpy(fecs[fec_n].rtp, fec, sizeof(ulp_fec_hdr_t) + fec_payload_len);
				
				fecs[fec_n].payload_len = fecs[fec_n].payload_original_len = fec_payload_len;
				fec_n++;
			}
		}
		
		switch_mutex_unlock(jb->mutex);
	}
	
#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "JB RED: found %u FECs for seq=%u level k=%u, jb=%p\n", fec_n, wanted_seq, k, (void*) jb);
#endif

	return fec_n;

jb_ulp_cb_find_fec_err:

	switch_mutex_unlock(jb->mutex);

	while (fec_n) {
		free(fecs[fec_n - 1].rtp);
		memset(&fecs[fec_n - 1], 0, sizeof(fec[0]));
		fec_n--;
	}

	return 0;
}

/** Receiver's callback called on RTP packet recovery. Need all packets of the level less 1. */
SWITCH_DECLARE(uint8_t) jb_ulp_cb_rtp_cover(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint8_t k, ulp_rtp_msg_t *rtp)
{
	return jb_ulp_rtp_material(policy, ntohs(fec->hdr.snb), k, rtp, 1);
}

/**
 * @seq - sequence successfully recovered which triggered this re recovery (making other recoveried possible)
 * @seq_shift - packets to check for possible recovery start from @seq - @seq_shift
 * @seq_n - number of packets to check for possible recovery (@seq_shift packets starting from @seq - @seq_shift)
 * @recursive_call_limit - max number of recursive calls
 * @ssrc - host ordered SSRC of the to-be-recovered-packet
 */
SWITCH_DECLARE(void) switch_jb_re_recover(ulp_policy_t *ulp_policy, uint16_t seq, uint16_t seq_shift, uint16_t seq_n, switch_rtp_packet_t *rerecovered_packet, uint32_t *recursive_call_limit, uint32_t *re_recovery_attempt_n, uint32_t ssrc)
{
	int i = 0;
	uint8_t missing = 0;

	switch_size_t recovered_len = 0;
	ulp_status_t ulp_err = ulp_status_fail;			/* after call to _recover contains additional information about recovery result */
	switch_jb_t *jb = ulp_get_user_data(ulp_policy);


	if (!ulp_policy || !jb || !rerecovered_packet || (*recursive_call_limit == 0)) {
		return;
	}
	*recursive_call_limit = *recursive_call_limit - 1;

#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_INFO, "JB RED: re-recovering based on seq %u, ulp=%p jb=%p (recursive_call_limit #%u)\n",
			seq, (void*) ulp_policy, (void*) ulp_get_user_data(ulp_policy), *recursive_call_limit);
#endif

	seq = seq > seq_shift ? seq - seq_shift : 0;

	/* Check and recover @seq_shift packets below and above @seq. */
	while (i < seq_n) {

		/**
		 * Check if it's missing and recover if yes.
		 */
		missing = 0;
		switch_mutex_lock(jb->mutex);
		if (switch_core_inthash_find(jb->missing_seq_hash, (uint32_t) htons(seq))) {
			missing = 1;
		} 

		switch_mutex_unlock(jb->mutex);

		if (missing) {
			int ret = 0;

			*re_recovery_attempt_n = *re_recovery_attempt_n + 1;

#if DEBUG_RED
			switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_INFO, "JB RED: re-recovering seq %u, ulp=%p jb=%p (re-recovery attempt #%u, recusive_call_limit #%u)\n",
					seq, (void*) ulp_policy, (void*) ulp_get_user_data(ulp_policy), *re_recovery_attempt_n, *recursive_call_limit);
#endif
			/* Recover packet. */
			memset((void*) rerecovered_packet, 0, sizeof(switch_rtp_packet_t));

			ret = switch_rtp_recover(ulp_policy, seq, &ulp_err, rerecovered_packet, ssrc);
			if (ulp_err == ulp_status_recovery_no_fecs) {
#if DEBUG_RED
				/* will mark a regular seq, when the actual FEC packet comes we'll take snb and mask16 and check if it's covered*/
				switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_DEBUG,
						"JB RED: marked seq [%d] for missing FEC . attempt_n %u\n", seq, *re_recovery_attempt_n);
#endif
				switch_core_inthash_insert(jb->missing_fecs_hash, (uint32_t) htons(seq), (void *) (intptr_t)switch_time_now());
			}
			recovered_len = (switch_size_t) ulp_rtp_last_recover_len(ulp_policy);

			/* Ignore partially recovered packets. */
			if (ret && (ulp_err == ulp_status_ok)) {

//#if DEBUG_RED
				switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_DEBUG, "JB RED: re-recovery [%s] seq=%u bytes=%zu\n", ulp_status_2_str[ulp_err], seq, recovered_len);
//#endif
				switch_jb_handle_recovered_pkt(jb, rerecovered_packet, recovered_len);
				
				/* Recover packets which are now possible to be recovered thanks to successful recovery of @seq. Limit recursion to what's available up to @recursive_call_limit. */
				switch_jb_re_recover(ulp_policy, seq, seq_shift, seq_n, rerecovered_packet, recursive_call_limit, re_recovery_attempt_n, ssrc);
			}
		}

		i++;
		seq++;
	}	
}

/**
 * NOTE: ULP policy must have FEC parsed already.
 */
uint8_t jb_recover_any_packets(ulp_fec_pkt_t *fec, ulp_policy_t *ulp_policy, uint32_t ssrc)
{
	int i = 0;
	uint8_t missing = 0;
	uint8_t forward = 0;
	uint8_t gk = 0;
	uint16_t mask16 = 0;
	uint32_t mask_cont = 0;

	uint16_t snb = ntohs(fec->hdr.snb);
	uint8_t lvl_n = 0;
	uint16_t seq = 0;
	uint32_t recursive_call_limit = SWITCH_RED_FEC_RE_RECOVER_RECURSIVE_CALL_N;
	uint32_t re_recovery_attempt_n = 0;
	
	switch_rtp_packet_t recovered_packet;			/* recovered packet */
	switch_size_t recovered_len = 0;
	ulp_status_t ulp_err = ulp_status_fail;			/* after call to _recover contains additional information about recovery result */
	uint8_t recovery_attempt_n = 0;
	switch_jb_t *jb = ulp_get_user_data(ulp_policy);


	if (!jb->missing_seq_hash) {
		return 0;
	}
	
#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_NOTICE, "JB RED: Triggering pre-recovering due to incoming FEC for seq=%u, ulp=%p jb=%p (recursive_calls=%u)\n",
			snb, (void*) ulp_policy, (void*) ulp_get_user_data(ulp_policy), recursive_call_limit);
#endif

	/* Recover packets which are now possible to be recovered thanks to new RTP. Limit recursion to SWITCH_RED_FEC_RE_RECOVER_MAX_N. */
	switch_jb_re_recover(ulp_policy, snb, SWITCH_RED_FEC_RE_RECOVER_SEQ_SHIFT, 2 * SWITCH_RED_FEC_RE_RECOVER_SEQ_SHIFT + 1, &recovered_packet, &recursive_call_limit, &re_recovery_attempt_n, ssrc);

#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_NOTICE, "JB RED: pre-recovering DONE for snb=%u (recursive_calls=%u, attempts=%u)\n",
			snb, SWITCH_RED_FEC_RE_RECOVER_RECURSIVE_CALL_N - recursive_call_limit, re_recovery_attempt_n);
#endif

	recursive_call_limit = SWITCH_RED_FEC_RE_RECOVER_RECURSIVE_CALL_N;
	re_recovery_attempt_n = 0;

	while (lvl_n < ulp_policy->lvl_n) {
	
		/* Process FEC levels: ULP policy has that info parsed from FEC already. */	
		gk = ulp_policy->lvl[lvl_n].gk;
		mask16 = ulp_policy->lvl[lvl_n].mask16;
		mask_cont = ulp_policy->lvl[lvl_n].mask_cont;
	
		i = 0;
		while (i < gk) {
			
			/**
			 * Get next seq protected by this FEC.
			 */
			seq = ulp_mask_2_seq(mask16, mask_cont, snb, i + 1);
			
#if DEBUG_RED
			switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_ALERT, "JB RED: found protected seq %u\n", seq);
#endif

			/**
			 * Check if it's missing and recover if yes.
			 */
			missing = 0;
			forward = 0;
			switch_mutex_lock(jb->mutex);
			if (switch_core_inthash_find(jb->missing_seq_hash, (uint32_t) htons(seq))) {
				missing = 1;
			} else if (!switch_core_inthash_find(jb->save_for_fec_hash, (uint32_t) htons(seq))) {
				forward = 1;
			}
			switch_mutex_unlock(jb->mutex);

			if (missing || forward) {
			
				recovery_attempt_n++;

#if DEBUG_RED
				switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_INFO, "JB RED: recovering (protected) seq %u, ulp=%p jb=%p\n", seq, (void*) ulp_policy, (void*) ulp_get_user_data(ulp_policy));
#endif
				/* Recover packet. */
				if (switch_rtp_recover(ulp_policy, seq, &ulp_err, &recovered_packet, ssrc)) {
					
					recovered_len = (switch_size_t) ulp_rtp_last_recover_len(ulp_policy);

#if DEBUG_RED
					switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_INFO, "JB RED: recovery [%s] (protected) seq=%u bytes=%zu\n", ulp_status_2_str[ulp_err], seq, recovered_len);
#endif
					/* Ignore partially recovered packets. */
					if (ulp_err == ulp_status_ok) {

						switch_jb_handle_recovered_pkt(jb, &recovered_packet, recovered_len);
					
//#if DEBUG_RED
						switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_DEBUG, "JB RED: Triggering re-recovering due to recovered seq %u, ulp=%p jb=%p (recursive_calls=%u)\n", seq, (void*) ulp_policy, (void*) ulp_get_user_data(ulp_policy), recursive_call_limit);
//#endif
						/* Recover packets which are now possible to be recovered thanks to successful recovery of @seq. Limit recursion to SWITCH_RED_FEC_RE_RECOVER_MAX_N. */
						switch_jb_re_recover(ulp_policy, seq, SWITCH_RED_FEC_RE_RECOVER_SEQ_SHIFT, 2 * SWITCH_RED_FEC_RE_RECOVER_SEQ_SHIFT + 1, &recovered_packet, &recursive_call_limit, &re_recovery_attempt_n, ssrc);
//#if DEBUG_RED
						switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_DEBUG, "JB RED: re-recovering DONE for seq=%u bytes=%zu (recursive_calls=%u, attempts=%u)\n",
								seq, recovered_len, SWITCH_RED_FEC_RE_RECOVER_RECURSIVE_CALL_N - recursive_call_limit, re_recovery_attempt_n);
//#endif
					}

				} else {
#if DEBUG_RED
					switch_log_printf(SWITCH_CHANNEL_SESSION_LOG(jb->session), SWITCH_LOG_INFO, "JB RED: recovery [%s] (protected) seq %u\n", ulp_status_2_str[ulp_err], seq);
#endif
				}
			}

			++i;
		}

		++lvl_n;
	}

	return recovery_attempt_n;
}

/**
 * Seq ordered list.
 *
 * Keep size of hashtable below SWITCH_RED_MAX_BUF_SZ and do not store sequences older than SWITCH_RED_MAX_SEQ_DIFF.
 */ 	
static void switch_jb_check_redundancy_hash(switch_jb_t *jb)
{
	switch_red_t *e = NULL;

	uint16_t seq = 0;
#if DEBUG_RED
	if (jb->redundancy_first && jb->redundancy_last) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ALERT, "JB RED: HASH entry first/last seq %u/%u from hash size=%u, jb=%p\n", ntohs(jb->redundancy_first->redundancy->header.seq), ntohs(jb->redundancy_last->redundancy->header.seq), jb->redundancy_hash_sz, (void*) jb);
	}
#endif

	/**
	 * Keep size of hahstable below the limit.
	 */
	if (jb->redundancy_hash_sz > SWITCH_RED_MAX_BUF_SZ) {

		/* Relink ordered list. */
		e = jb->redundancy_first;
		jb->redundancy_first = e->up;
		seq = e->redundancy->header.seq;
				
		switch_core_inthash_delete(jb->redundancy_hash, seq);
		jb->redundancy_hash_sz--;
		switch_safe_free(e->redundancy);
		switch_safe_free(e);

#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "JB RED: HASH removed first entry seq=%u from hash size=%u, jb=%p\n", ntohs(seq), jb->redundancy_hash_sz, (void*) jb);
#endif
	}

	while (jb->redundancy_first) {
		
		e = jb->redundancy_first;
		
		if (ntohs(e->redundancy->header.seq) + SWITCH_RED_MAX_SEQ_DIFF >= ntohs(jb->redundancy_last->redundancy->header.seq)) {
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ALERT, "JB RED: HASH break\n");
#endif
			break;
		}

#if DEBUG_RED
		seq = ntohs(e->redundancy->header.seq);
#endif
		jb->redundancy_first = e->up;

		e->up = NULL;
		switch_core_inthash_delete(jb->redundancy_hash, e->redundancy->header.seq);
		jb->redundancy_hash_sz--;
		switch_safe_free(e->redundancy);
		switch_safe_free(e);

#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ALERT, "JB RED: HASH removed entry seq=%u from hash size=%u, jb=%p\n", seq, jb->redundancy_hash_sz, (void*) jb);
#endif

	}
}

static void switch_jb_free_redundancy_hash_entries(switch_jb_t *jb)
{
	switch_red_t *e = NULL;
	
	if (jb->redundancy_last && jb->redundancy_first && 
			jb->redundancy_last->redundancy == jb->redundancy_first->redundancy) {
		/* this happens when we have only one redundancy packet in the hash - mainly with unit tests */
		switch_safe_free(jb->redundancy_last->redundancy);
		switch_safe_free(jb->redundancy_last);
		return;
	}
	while (jb->redundancy_first) {
		e = jb->redundancy_first;
		jb->redundancy_first = e->up;
		e->up = NULL;
		switch_safe_free(e->redundancy);
		switch_safe_free(e);
	}
}
/**
 * NOTE: ULP policy must have FEC parsed already (it is used in jb_recover_any_packets).
 */
SWITCH_DECLARE(switch_status_t) switch_jb_put_redundancy(switch_rtp_packet_t *redundancy, switch_size_t redundancy_len, ulp_policy_t *ulp_policy, uint16_t red_seq, uint32_t ssrc)
{
	switch_red_t *entry = NULL, *e = NULL;
	switch_jb_t *jb = ulp_get_user_data(ulp_policy);
	uint16_t base_seq, mask16;
	intptr_t then; int i = 0;
	void *outptr;
	uint32_t mask_cont;

	if (SWITCH_RED_MAX_BUF_SZ < 1) {
		return SWITCH_STATUS_FALSE;
	}
	if (!jb) {
		return SWITCH_STATUS_FALSE;
	}
	if (!jb->red_fec_seq_hash || !jb->redundancy_hash) {
		return SWITCH_STATUS_FALSE;
	}

	if (redundancy_len >= sizeof(switch_rtp_packet_t)) {
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "JB RED: trying to put %" SWITCH_SIZE_T_FMT " bytes exceeding buffer, truncate to %" SWITCH_SIZE_T_FMT "\n", redundancy_len, sizeof(switch_rtp_packet_t));
#endif
		redundancy_len = sizeof(switch_rtp_packet_t);
	}

	switch_mutex_lock(jb->mutex);
	jb->redundancy_count++;
	if (jb->missing_seq_hash && switch_core_inthash_delete(jb->missing_seq_hash, htons(red_seq))) {
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "JB RED: UNMARKED MISSING %u (FEC)\n", red_seq);
#endif
		jb_debug(jb, 2, "UNMARKED MISSING (FEC) %u\n", red_seq);
	}

	switch_malloc(entry, sizeof(switch_red_t));
	if (!entry) {
		switch_mutex_unlock(jb->mutex);
		return SWITCH_STATUS_FALSE;
	}
	memset(entry, 0, sizeof(switch_red_t));

	entry->redundancy = redundancy;										/* This is complete Redundant Encoding, e.g ULPFEC (header + payload), but not RED header. */
	entry->redundancy_len = redundancy_len;								/* Total length of the Redundant Encoding, e.g. ULPFEC packet. */
	entry->next = NULL;													/* Hash collisions for FECs with same snb. */
	entry->n = 1;														/* Bucket size (number of FECs with same snb in the bucket). */
	entry->up = NULL;													/* Seq ordered list */	
	
	/* Must MARK as RED-FEC before a call to switch_core_inthash_find, otherwise we may not MARK RED seq as FEC,
	 * because of multiple FECs for same snb (FEC's seq) but collisions not allowed. */
	switch_core_inthash_insert(jb->red_fec_seq_hash, (uint32_t) htons(red_seq), (void *)(intptr_t)1);

	/* Seq ordered list */	
	if (jb->redundancy_first == NULL) {
		jb->redundancy_first = entry;
	}

	e = switch_core_inthash_find(jb->redundancy_hash, redundancy->header.seq);
	if (e) {
		if (ulp_test_flag(ulp_policy, ULP_FL_ALLOW_FEC_COLLISION)) {

			/* Allow for collisions to be hashed to same key: multiple FECs with same snb but different protection masks. */
			entry->next = e;
			entry->n = e->n + 1;
		} else {
			switch_mutex_unlock(jb->mutex);
			return SWITCH_STATUS_SUCCESS;
		}
	}
	
	/**
	 * Sort the seq ordered list.
	 */	
	if (jb->redundancy_last) {
		jb->redundancy_last->up = entry;
	}
	jb->redundancy_last = entry;

	switch_core_inthash_insert(jb->redundancy_hash, redundancy->header.seq, (void *) entry);
	jb->redundancy_hash_sz++;

	switch_mutex_unlock(jb->mutex);
	base_seq = ntohs(redundancy->header.seq);
	mask16 = ntohs(((ulp_fec_pkt_t *)entry->redundancy)->lvl[0].mask16);
	mask_cont = ulp_lvl_protection_mask_cont(ulp_policy, 0);

	
#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "JB RED: HASH added entry seq=%u mask16=%.4x mask_cont=0x%x, bucket=%u, hash size=%u, jb=%p\n", base_seq, mask16, mask_cont, entry->n, jb->redundancy_hash_sz, (void*) jb);
#endif


	while (i < 16) {
		uint32_t seq = ulp_mask_2_seq(mask16, mask_cont, base_seq, i + 1);

		if ((outptr = switch_core_inthash_find(jb->missing_fecs_hash, htons(seq)))) {
			then = (intptr_t) outptr;
			jb->late_fec_count++;
			jb->avg_late_fec = (uint32_t)(switch_time_now() - then);
#if DEBUG_RED
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, 
					"JB RED: received late FEC packet. base seq: [%d] seq: [%d] count[%d] time_diff [%uus]\n",
					base_seq, seq, (uint32_t)jb->late_fec_count, jb->avg_late_fec);
#endif
			switch_core_inthash_delete(jb->missing_fecs_hash, htons(seq));
		}
		i++;
	}


	/**
	 * Attempt any recovery possible using this FEC.
	 */
	jb_recover_any_packets((ulp_fec_pkt_t *) redundancy, ulp_policy, ssrc);
	
	switch_mutex_lock(jb->mutex);
	
	/**
	 * Keep control over the size of seq ordered list - remove seq which fall to far away from current seq.
	 */	
	switch_jb_check_redundancy_hash(jb);

	switch_mutex_unlock(jb->mutex);

	return SWITCH_STATUS_SUCCESS;
}

static void switch_jb_reset_redundancy_hash(switch_jb_t *jb)
{
	if (jb->redundancy_hash) {
		switch_jb_free_redundancy_hash_entries(jb);
		switch_core_inthash_destroy(&jb->redundancy_hash);
		switch_core_inthash_init(&jb->redundancy_hash);
		jb->redundancy_hash_sz = 0;
		jb->redundancy_first = NULL;
		jb->redundancy_last = NULL;
	}
	if (jb->red_fec_seq_hash) {
		switch_core_inthash_destroy(&jb->red_fec_seq_hash);
		switch_core_inthash_init(&jb->red_fec_seq_hash);
	}
}

SWITCH_DECLARE(void) switch_jb_set_ulp_policy(switch_jb_t *jb, ulp_policy_t *policy)
{
	if (!jb) {
		return;
	}

	jb->ulp_policy = policy;
}

SWITCH_DECLARE(ulp_policy_t *) switch_jb_get_ulp_policy(switch_jb_t *jb)
{
	if (!jb) {
		return NULL;
	}

	return jb->ulp_policy;
}

SWITCH_DECLARE(void) switch_jb_handle_recovered_pkt(switch_jb_t *jb, switch_rtp_packet_t *recovered_packet, switch_size_t recovered_len)
{

#ifdef TEST_ULPFEC

	if (switch_jb_put_rtp_recovered(jb, recovered_packet, recovered_len - 12, 0) != SWITCH_STATUS_SUCCESS) {

		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED; TEST: failed inserting into recovered packets hash\n");
		return;
	}

#else // if not TEST_ULPFEC

	switch_jb_put_packet(jb, recovered_packet, recovered_len, 1);
#ifdef DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "JB RED: deleting RECOVERED seq=%u %u from missing seq hash\n", recovered_packet->header.seq, ntohs(recovered_packet->header.seq));
#endif

	/** switch_jb_put_packet removed seq from missing hash, but prerform just in case check. */ 
	if (switch_core_inthash_delete(jb->missing_seq_hash, (uint32_t) recovered_packet->header.seq)) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "JB RED: deleted RECOVERED seq=%u from missing seq hash\n", ntohs(recovered_packet->header.seq));
	}
#endif
}

SWITCH_DECLARE(int) switch_jb_check_seq_is_red_fec(switch_jb_t *jb, uint16_t seq)
{
	return (switch_core_inthash_find(jb->red_fec_seq_hash, (uint32_t) htons(seq)) ? 1 : 0);
}

SWITCH_DECLARE(int) switch_jb_check_seq_is_fec(switch_jb_t *jb, uint16_t seq)
{
	return (switch_core_inthash_find(jb->redundancy_hash, htons(seq)) ? 1 : 0);
}

#ifdef TEST_ULPFEC

static switch_status_t switch_jb_put_rtp_recovered(switch_jb_t *jb, switch_rtp_packet_t *rtp, switch_size_t rtp_len, uint8_t is_red)
{
	switch_rtp_for_fec_t *entry = NULL;

	if (SWITCH_RTP_FOR_FEC_MAX_BUF_SZ < 1) {
		return SWITCH_STATUS_FALSE;
	}

	if (rtp_len >= sizeof(switch_rtp_packet_t)) {
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "RTP FEC RED: trying to put %" SWITCH_SIZE_T_FMT " bytes exceeding buffer, truncate to %" SWITCH_SIZE_T_FMT "\n", rtp_len, sizeof(switch_rtp_packet_t));
#endif
		rtp_len = sizeof(switch_rtp_packet_t);
	}

	entry = switch_core_alloc(jb->pool, sizeof(switch_rtp_for_fec_t));
	if (!entry) {
		return SWITCH_STATUS_FALSE;
	}
	memset(entry, 0, sizeof(switch_rtp_for_fec_t));

	/* Skip RED Encoding header (1 byte). */
	memcpy(&entry->rtp, rtp, sizeof(switch_rtp_hdr_t));					/* Copy RTP header. */
	memcpy(&entry->rtp.body, (char*)rtp + (is_red ? 13 : 12), rtp_len);	/* Copy RTP payload. */
	entry->rtp_len = rtp_len;											/* Payload length. */
	entry->next = NULL;													/* Hash collisions with same sequence. */
	entry->n = 1;														/* Bucket size (number of RTP with same sequence in the bucket). */
	entry->up = NULL;													/* Seq ordered list */	


	/**
	 * Insert into hash by network-ordered seq.
	 */
	switch_core_inthash_insert(jb->rtp_recovered_hash, rtp->header.seq, (void *) entry);
	jb->rtp_recovered_count++;

#if DEBUG_RED
	seq = rtp->header.seq;
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RTP FEC RED: HASH added recovered entry seq=%u, bucket size=%u, hash size=%u\n", ntohs(seq), entry->n, jb->rtp_recovered_hash_sz);
#endif

	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t switch_jb_test_recovered(switch_jb_t *jb, switch_rtp_packet_t *rtp_nack, switch_rtp_packet_t *rtp_fec, switch_size_t rtp_nack_len, switch_size_t rtp_fec_len)
{
	uint16_t seq = rtp_nack->header.seq;
	switch_size_t payload_bytes = 0;

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "RTP FEC RED: TEST: Testing recovery for seq=%u (FEC bytes=%zu, NACK bytes=%zu)\n", ntohs(seq), rtp_fec_len, rtp_nack_len);

	/* Test recovery of header. */

	if (rtp_fec_len != rtp_nack_len) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED: TEST: Failed length check for seq=%d\n", ntohs(seq));
	}
	if (!memcmp(rtp_nack, rtp_fec, 12)) {
	
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "RTP FEC RED: TEST: Header recovery for seq=%u OK\n", ntohs(seq));

	} else {
	
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED: TEST: Header recovery for seq=%u FAIL\n", ntohs(seq));
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED: TEST: FEC-recovered:\n");
		switch_rtp_print_header((switch_rtp_hdr_t *) rtp_fec);
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED: TEST: NACK-recovered:\n");
		switch_rtp_print_header((switch_rtp_hdr_t *) rtp_nack);
	}
	
	payload_bytes = ulp_min(rtp_nack_len, rtp_fec_len) - 12; /*compare only payloads*/

	if (!memcmp(rtp_nack->body, rtp_fec->body, payload_bytes)) {
	
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "RTP FEC RED: TEST: payload recovery for seq=%u OK\n", ntohs(seq));

	} else {

		int diff = ulp_test_bytes(rtp_nack->body, rtp_fec->body, payload_bytes);
	
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED: TEST: Payload recovery for seq=%u FAIL (bytes=%zu), different byte is #%d 0x%X vs 0x%X\n", ntohs(seq), payload_bytes, diff, *((char*)rtp_fec->body + diff) & 0xFF, *((char*)rtp_nack->body + diff) & 0xFF);
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED: TEST: FEC-recovered:\n");
		switch_rtp_dump_bytes((char*) rtp_fec->body, 8);
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "RTP FEC RED: TEST: NACK-recovered:\n");
		switch_rtp_dump_bytes((char *) rtp_nack->body, 8);
	}

	return SWITCH_STATUS_FALSE;
}

#endif // TEST_ULPFEC


/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 noet:
 */
